function Set-ClientCredentials 
{
    param (
        [Parameter(Mandatory=$true)]
        [Object]$wcfClient,

        [Parameter(Mandatory=$true)]
        [string]$cmsAuth,

        [Parameter(Mandatory=$false)]
        [string]$cmsUserName,

        [Parameter(Mandatory=$false)]
        [string]$cmsUserPassword
    )

    switch ($cmsAuth)
    {
        "Windows"{
            if ($cmsUserName)
            {
                Write-Verbose "Using Windows authentication with CMS user name: $cmsUserName"
                $wcfClient.ClientCredentials.Windows.ClientCredential = New-Object System.Net.NetworkCredential $cmsUserName, $cmsUserPassword
            }
            else
            {
                Write-Verbose "Using Windows authentication with the current Windows user's credentials (no CMS user specified explicitly)." 
                $wcfClient.ClientCredentials.Windows.ClientCredential = [System.Net.CredentialCache]::DefaultNetworkCredentials;
            }
        }
        "Basic" {
            if (!$cmsUserName -or !$cmsUserPassword)
            {
                throw "CMS user name and password must be specified for Basic authentication."
            }
            Write-Verbose "Using Basic authentication with CMS user name: $cmsUserName"
            $wcfClient.ClientCredentials.UserName.UserName = $cmsUserName
            $wcfClient.ClientCredentials.UserName.Password = $cmsUserPassword
        }
    }
}

function Get-CoreServiceClient
{
    param (
        [Parameter(Mandatory=$true)]
        [string]$coreSeviceClientUrl,

        [Parameter(Mandatory=$true)]
        [ValidateSet("Windows","Basic")]
        [string]$authType,

        [Parameter(Mandatory=$false)]
        [string]$userName,

        [Parameter(Mandatory=$false)]
        [string]$password
    )        
    Write-Host "Getting Core Service Client with type '$authType' for CMS URL '$coreSeviceClientUrl' ..."

	$binding = New-Object System.ServiceModel.BasicHttpBinding
	$binding.MaxBufferPoolSize = [int]::MaxValue
	$binding.MaxReceivedMessageSize = [int]::MaxValue
	$binding.ReaderQuotas.MaxArrayLength = [int]::MaxValue
	$binding.ReaderQuotas.MaxBytesPerRead = [int]::MaxValue
	$binding.ReaderQuotas.MaxStringContentLength = [int]::MaxValue
	$binding.ReaderQuotas.MaxNameTableCharCount = [int]::MaxValue
	
    if ($coreSeviceClientUrl.StartsWith("https")) { $binding.Security.Mode = "Transport" }
    else { $binding.Security.Mode = "TransportCredentialOnly" }
    $binding.Security.Transport.ClientCredentialType = $authType
    $endpoint = New-Object System.ServiceModel.EndpointAddress ($coreSeviceClientUrl + "webservices/CoreService201701.svc/basicHttp")
    $client = New-Object Tridion.ContentManager.CoreService.Client.CoreServiceClient $binding,$endpoint
    
    Set-ClientCredentials $client $authType $userName $password

    Write-Host "Endpoint address: $($client.Endpoint.Address)"

    if ($type -eq "Service")
    {
        $cmsUser = $client.GetCurrentUser()
        $cmsUserTitle = $cmsUser.Title
        $cmsUserDescription = $cmsUser.Description
        Write-Host "Connected to Core Service as user '$cmsUserTitle' ($cmsUserDescription)" 
    }
    return $client;
}

#=============================================
Add-Type -assemblyName mscorlib
Add-Type -assemblyName System.ServiceModel
$ScriptDir = Split-Path -parent $MyInvocation.MyCommand.Path
Add-Type -Path $ScriptDir\lib\Tridion.ContentManager.CoreService.Client.dll