function Get-PagesByPageTemplate
{
    param (
        [Parameter(Mandatory=$true)]
        [Object]$coreServiceClient,

        [Parameter(Mandatory=$true)]
        [string]$pageTemplateUrl
    )

    function Get-PublicationIdsFromOrgStructure($pageTemplateId, $client)
    {
        $linkToRepository = New-Object "Tridion.ContentManager.CoreService.Client.LinkToRepositoryLocalObjectData"
        $linkToRepository.IdRef = $pageTemplateId
        $filter = New-Object "Tridion.ContentManager.CoreService.Client.BluePrintNodesFilterData"
        $filter.ForItem = $linkToRepository
    
        $result = @($client.GetSystemWideList($filter) `
            | Where-Object Item -ne $null `
            | Select-Object -ExpandProperty Id)
        
        return ,$result
    }

    function Get-PagesFromPublications($publicationIds, $pageTemplateId, $client)
    {
        $pageTemplateItemType = [Tridion.ContentManager.ItemType]::PageTemplate
        [Tridion.ContentManager.CoreService.Client.PageData[]]$usedPages = @()
        foreach ($publicationId in $publicationIds)
        {
            $publication = $client.Read($publicationId, $null)
            $currentPublicationUri = New-TcmUri $publicationId
            $pageTemplateTcmURi = New-TcmUri $pageTemplateId
            $currentPageTemplateUri = (New-Object "Tridion.ContentManager.TcmUri" -ArgumentList $pageTemplateTcmURi.ItemId, $pageTemplateItemType, $currentPublicationUri.ItemId).ToString()
            $pages = Get-Pages $publication $client
            foreach ($pageItem in $pages)
            {
                $page = $pageItem.Item
                if ($page.PageTemplate.IdRef -eq $currentPageTemplateUri)
                {
                    $usedPages += $page
                }
            }
        }
        return ,$usedPages
    }

    Write-InfoMessage "Getting local and localized Pages with Page Template '$pageTemplateUrl'... "
    
    $publicationIds = Get-PublicationIdsFromOrgStructure $pageTemplateUrl $coreServiceClient
    $pages = Get-PagesFromPublications $publicationIds $pageTemplateUrl $coreServiceClient

    Write-InfoMessage "Done. Loaded $($pages.Length) page(s)."
    return $pages
}

function New-TcmUri
{
    param (
        [Parameter(Mandatory=$true)]
        [string]$tcmUri
    )
    return New-Object "Tridion.ContentManager.TcmUri" $tcmUri
}

function Get-TcmUriEqualsWithoutPublicationId
{
    param (
        [Parameter(Mandatory=$true)]
        [string]$tcmUriA,

        [Parameter(Mandatory=$true)]
        [string]$tcmUriB
    )

    $tcmUriObjA = New-TcmUri $tcmUriA
    $tcmUriObjB = New-TcmUri $tcmUriB
    return (($tcmUriObjA.ItemId -eq $tcmUriObjB.ItemId) -and ($tcmUriObjA.ItemType -eq $tcmUriObjB.ItemType))
}

function Create-TcmUriForPublication
{
    param (
        [Parameter(Mandatory=$true)]
        [string]$tcmUriWithOldPublication,

        [Parameter(Mandatory=$true)]
        [string]$tcmUriWithNewPublication
    )

    $tcmUriObjWithOldPublication = New-TcmUri $tcmUriWithOldPublication
    $tcmUriObjWithNewPublication = New-TcmUri $tcmUriWithNewPublication
    $tcmUri = New-Object "Tridion.ContentManager.TcmUri" -ArgumentList $tcmUriObjWithOldPublication.ItemId,$tcmUriObjWithOldPublication.ItemType,$tcmUriObjWithNewPublication.PublicationId
    return $tcmUri
}

function Set-PageTemplate
{
    param (
        [Parameter(Mandatory=$true)]
        [Object]$coreServiceClient,

        [Parameter(Mandatory=$true)]
        [Tridion.ContentManager.CoreService.Client.PageData]$page,

        [Parameter(Mandatory=$true)]
        [Object]$pageTemplate
    )

    Write-InfoMessage "Updating Page Template from '$($page.PageTemplate.IdRef)' to '$($pageTemplate.Id)'..."
    $currentPageTemplateTcmUri = $page.PageTemplate.IdRef
    $newPageTemplateTcmUri = $pageTemplate.Id
    if ((Get-TcmUriEqualsWithoutPublicationId $currentPageTemplateTcmUri $newPageTemplateTcmUri) -eq $false)
    {
        $pageTemplateTcmURi = Create-TcmUriForPublication $newPageTemplateTcmUri $currentPageTemplateTcmUri
        $pageTemplate = New-Object "Tridion.ContentManager.CoreService.Client.LinkToPageTemplateData"
        $pageTemplate.IdRef = $pageTemplateTcmURi.ToString()
        $page.PageTemplate = $pageTemplate
        $page.IsPageTemplateInherited = $false
    }
    Write-InfoMessage "Done"
}

function Get-RegionNameFromComponentTemplate
{
    param (
        [Parameter(Mandatory=$true)]
        [Object]$componentTemplate
    )

    function StripModuleName([string]$qualifiedName)
    {
        if ([string]::IsNullOrEmpty($qualifiedName))
        {
            return $null
        }
        $parts = $qualifiedName.Split(':')
        $viewName = switch ($parts.Count)
        {
            1 {$parts[0].Trim()}
            2 {$parts[1].Trim()}
            default {$null}
        }
        return $viewName
    }

    [string]$regionName = $null
    if ($componentTemplate.Metadata)
    {
        $metadataXml = [xml]$componentTemplate.Metadata
        $regionName = $metadataXml.Metadata.regionName
        if ([string]::IsNullOrEmpty($regionName))
        {
            $regionName = StripModuleName $metadataXml.Metadata.regionView
        }
    }
    Write-InfoMessage "Region Name: '$($regionName)'"
    return $regionName
}

function Get-RegionNameFromComponentPresentation
{
    param (
        [Parameter(Mandatory=$true)]
        [Object]$coreServiceClient,

        [Parameter(Mandatory=$true)]
        [Object]$componentPresentation
    )

    $componentTemplateId = $componentPresentation.ComponentTemplate.IdRef
    if ($componentTemplateId -ne [Tridion.ContentManager.TcmUri]::UriNull.ToString())
    {
        $componentTemplate = $coreServiceClient.Read($componentTemplateId, $null)
        if ($componentTemplate)
        {
            Write-InfoMessage "Getting Region from Component Template '$($componentTemplateId)'..."
            $regionName = Get-RegionNameFromComponentTemplate $componentTemplate
            Write-InfoMessage "Done"
            return $regionName
        }
    }
    return $null
}

function Get-ComponentPresentationsToRegionNamesMap
{
    param (
        [Parameter(Mandatory=$true)]
        [Object]$coreServiceClient,

        [Parameter(Mandatory=$true)]
        [Object]$componentPresentations
    )

    Write-InfoMessage "Creating the mapper of Component Presenation to Region Name..."
    $componentPresentationsToRegionNamesMap = [ordered]@{}
    foreach ($componentPresentation in $componentPresentations)
    {
        $regionName = Get-RegionNameFromComponentPresentation $coreServiceClient $componentPresentation
        $componentPresentationsToRegionNamesMap.Add($componentPresentation, $regionName)
    }
    Write-InfoMessage "Done"
    return $componentPresentationsToRegionNamesMap
}

function Set-NewComponentTemplateInComponentPresentation
{
    param (
        [Parameter(Mandatory=$false)]
        [Object]$componentPresentations,

        [Parameter(Mandatory=$true)]
        [Hashtable]$componentTemplateMapper
    )

    if (!$componentPresentations) 
    {
        return
    }
    foreach ($componentPresentation in $componentPresentations)
    {
        $componentTemplateId = $componentPresentation.ComponentTemplate.IdRef
        $newComponentTemplateIdKey = $componentTemplateMapper.Keys.GetEnumerator() | Where-Object { Get-TcmUriEqualsWithoutPublicationId $componentTemplateId $_ } | Select-Object $_ -First 1
        if ($newComponentTemplateIdKey)
        {
            $tcmUriObj = Create-TcmUriForPublication $componentTemplateMapper[$newComponentTemplateIdKey] $componentTemplateId
            $tcmUri = $tcmUriObj.ToString()
            Write-InfoMessage "Updating Component Template from '$($componentTemplateId)' to '$($tcmUri)'..."
            $componentTemplateLink = New-Object "Tridion.ContentManager.CoreService.Client.LinkToComponentTemplateData"
            $componentTemplateLink.IdRef = $tcmUri
            $componentPresentation.ComponentTemplate = $componentTemplateLink
            Write-InfoMessage "Done"
        }
    }
}

function Set-NewComponentTemplateInComponentPresentationForRegions
{
    param (
        [Parameter(Mandatory=$true)]
        [Object]$regions,

        [Parameter(Mandatory=$true)]
        [Hashtable]$componentTemplateMapper
    )

    foreach ($region in $regions)
    {
        Write-InfoMessage "Processing Region '$($region.RegionName)'..."
        Set-NewComponentTemplateInComponentPresentation $region.ComponentPresentations $componentTemplateMapper
        Set-NewComponentTemplateInComponentPresentationForRegions $region.Regions $componentTemplateMapper
        Write-InfoMessage "Done"
    }
}

function Set-NewComponentTemplate
{
    param (
        [Parameter(Mandatory=$true)]
        [Object]$page,

        [Parameter(Mandatory=$true)]
        [Hashtable]$componentTemplateMapper
    )

    Write-InfoMessage "Processing Page '$($page.Title)'..."
    Set-NewComponentTemplateInComponentPresentation $page.ComponentPresentations $componentTemplateMapper
    Set-NewComponentTemplateInComponentPresentationForRegions $page.Regions $componentTemplateMapper
    Write-InfoMessage "Done"
}

function Set-PageRegions
{
    param (
        [Parameter(Mandatory=$true)]
        [Object]$page,

        [Parameter(Mandatory=$true)]
        [Object]$coreServiceClient
    )
    
    Write-InfoMessage "Adding Regions on the page..."
    $pageTemplate = $client.Read($page.PageTemplate.IdRef, $null)
    $regionSchemaId = $pageTemplate.PageSchema.IdRef
    Set-NestedRegions $page $regionSchemaId
    Write-InfoMessage "Done"
}

function Set-NestedRegions
{
    param (
        [Parameter(Mandatory=$true)]
        [Object]$parentRegion,

        [Parameter(Mandatory=$true)]
        [string]$regionSchemaId
    )


    if ($regionSchemaId -eq [Tridion.ContentManager.TcmUri]::UriNull.ToString())
    {
        Write-WarningMessage "Changing of Region Schema is skipped. The Region Schema is not specified."
        return
    }

    $regionSchema = $client.Read($regionSchemaId, $null)
    ForEach ($nestedRegion in $regionSchema.RegionDefinition.NestedRegions) 
    {
        Write-InfoMessage "Processing the region '$($nestedRegion.RegionName)'..."
        $currentRegion = $parentRegion.Regions | Where-Object {$_.RegionName -eq $nestedRegion.RegionName}
        if ($null -eq $currentRegion) 
        {
            Write-InfoMessage "The Region does not exist. Creating a new one..."
            $currentRegion = New-Object "Tridion.ContentManager.CoreService.Client.EmbeddedRegionData"
            $currentRegion.RegionName = $nestedRegion.RegionName
            $currentRegion.RegionSchema = $nestedRegion.RegionSchema
            $currentRegion.Regions = New-Object "Tridion.ContentManager.CoreService.Client.RegionList"
            $parentRegion.Regions.Add($currentRegion)
            Write-InfoMessage "Done"
        }
        elseif ($null -eq $currentRegion.Regions)
        {
            $currentRegion.Regions = New-Object "Tridion.ContentManager.CoreService.Client.RegionList"
        }
        Write-InfoMessage "Done"
        Write-InfoMessage "Processing the region '$($currentRegion.RegionName)'..."
        Set-NestedRegions $currentRegion $nestedRegion.RegionSchema.IdRef
        Write-InfoMessage "Done"
    }
}

function Move-PageComponentPresentations
{
    param (
        [Parameter(Mandatory=$true)]
        [Object]$page,

        [Parameter(Mandatory=$true)]
        [System.Object]$componentPresenation,

        [Parameter(Mandatory=$true)]
        [string]$pageTemplateId,

        [Parameter(Mandatory=$true)]
        [AllowEmptyString()]
        [string]$sourceRegionName,

        [Parameter(Mandatory=$true)]
        [Object]$client
    )
    
    $key = $pageTemplateId + "#" + $sourceRegionName
    $regionInfo = $ItemMapperStorage["ComponentPresentationDestinationRegionMap"][$key]
    $destinationRegionName = $regionInfo.RegionName
    if ($null -eq $destinationRegionName)
    {
        $destinationRegionName = $sourceRegionName
    }

    $destinationRegion = Get-RegionByFullyQualifiedName $page $destinationRegionName
    if ($null -ne $regionInfo.Metadata -and $null -ne $destinationRegion)
    {
        $regionSchema = $client.Read($destinationRegion.RegionSchema.IdRef, $null)
        $metadataNamespaceUri = $regionSchema.NamespaceUri
        $metadata = ConvertTo-TcmXml $regionInfo.Metadata -namespaceAttributeName "xmlns" -namespaceAttributeValue $metadataNamespaceUri
        $destinationRegion.Metadata = $metadata.InnerXml
    }

    if ((Add-RegionComponentPresentation $page $componentPresenation $destinationRegion) -eq $true)
    {
        Remove-PageComponentPresentation $page $componentPresenation
    }
}

function Add-RegionComponentPresentation
{ 
    param (
        [Parameter(Mandatory=$true)]
        [Object]$page,

        [Parameter(Mandatory=$true)]
        [Object]$componentPresenation,
        
        [Parameter(Mandatory=$false)]
        [Object]$destinationRegion
    )

    Write-InfoMessage "Adding Component Presentation (Component: '$($componentPresenation.Component.Title)' ($($componentPresenation.Component.IdRef)) and Component Template '$($componentPresenation.ComponentTemplate.Title)' ($($componentPresenation.ComponentTemplate.IdRef))) to the Region '$destinationRegionName'..."
    if ($null -ne $destinationRegion)
    {
        $regionComponentPresentations = $destinationRegion.ComponentPresentations
        if ($null -eq $regionComponentPresentations)
        {
            [Tridion.ContentManager.CoreService.Client.ComponentPresentationData[]]$regionComponentPresentations = @()
        }
        $regionComponentPresentations += $componentPresenation
        $destinationRegion.ComponentPresentations = $regionComponentPresentations
        Write-InfoMessage "Done"
        return $true
    }
    Write-WarningMessage "Skipped"
    return $false
}

function Remove-PageComponentPresentation
{
    param (
        [Parameter(Mandatory=$true)]
        [Object]$page,

        [Parameter(Mandatory=$true)]
        [System.Object]$componentPresenation
    )

    Write-InfoMessage "Removing Component Presentation (Component: '$($componentPresenation.Component.Title)' ($($componentPresenation.Component.IdRef)) and Component Template '$($componentPresenation.ComponentTemplate.Title)' ($($componentPresenation.ComponentTemplate.IdRef))) from the Page..."
    [Tridion.ContentManager.CoreService.Client.ComponentPresentationData[]]$pageComponentPresentations = $page.ComponentPresentations
    $pageComponentPresentations = 0..($pageComponentPresentations.Length-1) | 
        ForEach-Object { if ($PSItem -ne $pageComponentPresentations.IndexOf($componentPresenation)) { $pageComponentPresentations[$PSItem] } }
    
    if ($null -ne $pageComponentPresentations) 
    {
        $page.ComponentPresentations = $pageComponentPresentations
    }
    else
    {
        [Tridion.ContentManager.CoreService.Client.ComponentPresentationData[]]$emptyComponentPresentations = @()
        $page.ComponentPresentations = $emptyComponentPresentations
    }
    Write-InfoMessage "Done"
}

function Get-RegionByFullyQualifiedName
{
    param (
        [Parameter(Mandatory=$true)]
        [Object]$parentRegion,

        [Parameter(Mandatory=$true)]
        [AllowEmptyString()]
        [string]$regionFullyQualifiedName
    )

    Write-InfoMessage "Searching for the Region with fully-qualified name '$regionFullyQualifiedName'..."
    if ([string]::IsNullOrEmpty($regionFullyQualifiedName) -or $regionFullyQualifiedName -eq "\")
    {
        Write-InfoMessage "Done"
        return $parentRegion
    }

    $foundRegion = New-Object "Tridion.ContentManager.CoreService.Client.EmbeddedRegionData"
    $regionNames = $regionFullyQualifiedName.split("\")
    for ($i = 0; $i -lt $regionNames.length; $i++)
    {
        $regionName = $regionNames[$i]
        [Tridion.ContentManager.CoreService.Client.RegionList]$nestedRegions = $parentRegion.Regions
        $foundRegion = $parentRegion.Regions | Where-Object {$_.RegionName -eq $regionName}
        if ($null -ne $foundRegion)
        {
            $parentRegion = $foundRegion
        }
        else 
        {
            Write-WarningMessage "Skipped. Can't find the region."
            return $null
        }
    }
    Write-InfoMessage "Done"
    return $foundRegion
}

function Get-Pages
{
    param (
        [Parameter(Mandatory=$true)]
        [Object]$publication,

        [Parameter(Mandatory=$true)]
        [Object]$client
    )

    [Tridion.ContentManager.CoreService.Client.ItemType[]]$itemTypes = @()
    $itemTypes += [Tridion.ContentManager.CoreService.Client.ItemType]::Page
    $filter = New-Object "Tridion.ContentManager.CoreService.Client.OrganizationalItemItemsFilterData"
    $filter.ItemTypes = $itemTypes
    $filter.BaseColumns = [Tridion.ContentManager.CoreService.Client.ListBaseColumns]::Extended
    $filter.Recursive = $true

    $publicationPages = $client.GetList($publication.RootStructureGroup.IdRef, $filter)

    $localPagesIds = $publicationPages.GetEnumerator()`
        | Where-Object {($_ -is [Tridion.ContentManager.CoreService.Client.PageData])}`
        | Where-Object {($_.BluePrintInfo.IsShared -eq $false -or $_.BluePrintInfo.IsLocalized -eq $true)}`
        | Select-Object -ExpandProperty Id

    $pages = @()
    if ($null -ne $localPagesIds)
    {
        $readPages = $client.BulkRead($localPagesIds, $null)
        $pages = $readPages.Values.GetEnumerator()`
        | Select-Object ($_.Item)`
    }

    return ,$pages
}

function ConvertTo-TcmXml
{
    Param(
        [Parameter(ValueFromPipeline = $true, Position = 0)]
        [System.Collections.Specialized.OrderedDictionary]$inputObject,

        [System.String]$root="Metadata",

        [System.String]$namespaceAttributeName,
        [System.String]$namespaceAttributeValue
    )

    function ScriptBlock 
    {
        Param($elem, $root)
        if( $elem.Value -is [System.Collections.Specialized.OrderedDictionary] ){
            $rootNode = $root.AppendChild($Doc.CreateNode([System.Xml.XmlNodeType]::Element,$elem.Key,$Null))
            $elem.Value.GetEnumerator() | ForEach-Object {
                Scriptblock $_ $rootNode
            }
        }
        else{
            $element = $doc.CreateElement($elem.Key)
            $element.InnerText = if($elem.Value -is [Array]) {
                $elem.Value -join ','
            }
            else{
                $elem.Value.ToString()
            }
            $root.AppendChild($element) | Out-Null
        }
    }

    $doc = [xml]"<$($root)></$($root)>"
    $inputObject.GetEnumerator() | ForEach-Object {
        ScriptBlock $_ $doc.DocumentElement
    }

    $doc.DocumentElement.SetAttribute($namespaceAttributeName, $namespaceAttributeValue);
    return $doc 
}

#=============================================
$ScriptDir = Split-Path -parent $MyInvocation.MyCommand.Path
Add-Type -Path $ScriptDir\lib\Tridion.Common.dll
